@@ -1,7 +1,7 @@
 #
 # api.py
 #
-# Copyright (C) 2017-2021 frnmst (Franco Masotti) <franco.masotti@tutanota.com>
+# Copyright (C) 2017-2022 Franco Masotti (franco \D\o\T masotti {-A-T-} tutanota \D\o\T com)
 #
 # This file is part of md-toc.
 #
@@ -21,16 +21,17 @@
 """The main file."""
 
 import copy
+import os
 import re
 import sys
 
 import fpyutils
 
-from . import cmark, generic
+from . import generic
+from .cmark import inlines_c, node_h, references_h
 from .constants import common_defaults
 from .constants import parser as md_parser
-from .exceptions import (CannotTreatUnicodeString, GithubEmptyLinkLabel,
-                         GithubOverflowCharsLinkLabel,
+from .exceptions import (GithubEmptyLinkLabel, GithubOverflowCharsLinkLabel,
                          GithubOverflowOrderedListMarker,
                          StdinIsNotAFileToBeWritten,
                          StringCannotContainNewlines,
@@ -38,16 +39,19 @@ from .exceptions import (CannotTreatUnicodeString, GithubEmptyLinkLabel,
 
 
 def write_string_on_file_between_markers(filename: str, string: str,
-                                         marker: str):
+                                         marker: str, newline_string: str = common_defaults['newline string']):
     r"""Write the table of contents on a single file.
 
     :parameter filename: the file that needs to be read or modified.
     :parameter string: the string that will be written on the file.
     :parameter marker: a marker that will identify the start
          and the end of the string.
+    :parameter newline_string: the new line separator.
+         Defaults to ``os.linesep``.
     :type filenames: str
     :type string: str
     :type marker: str
+    :type newline_string: str
     :returns: None
     :rtype: None
     :raises: StdinIsNotAFileToBeWritten or an fpyutils exception
@@ -74,11 +78,12 @@ def write_string_on_file_between_markers(filename: str, string: str,
             final_string,
             marker_line_positions[1],
             filename,
-            append=False)
+            append=False,
+            newline_character=newline_string)
 
 
 def write_strings_on_files_between_markers(filenames: list, strings: list,
-                                           marker: str):
+                                           marker: str, newline_string: str = common_defaults['newline string']):
     r"""Write the table of contents on multiple files.
 
     :parameter filenames: the files that needs to be read or modified.
@@ -86,9 +91,12 @@ def write_strings_on_files_between_markers(filenames: list, strings: list,
          string is associated with one file.
     :parameter marker: a marker that will identify the start
          and the end of the string.
+    :parameter newline_string: the new line separator.
+         Defaults to ``os.linesep``.
     :type filenames: list
-    :type string: list
+    :type strings: list
     :type marker: str
+    :type newline_string: str
     :returns: None
     :rtype: None
     :raises: an fpyutils exception or a built-in exception.
@@ -106,7 +114,7 @@ def write_strings_on_files_between_markers(filenames: list, strings: list,
 
     file_id = 0
     for f in filenames:
-        write_string_on_file_between_markers(f, strings[file_id], marker)
+        write_string_on_file_between_markers(f, strings[file_id], marker, newline_string)
         file_id += 1
 
 
@@ -143,8 +151,8 @@ def build_toc(filename: str,
          Defaults to ``0```.
     :parameter constant_ordered_list: use a single integer
         as list marker. This sets ordered to ``True``.
-    :parameter newline_string: the set of characters used to
-        go to a new line.
+    :parameter newline_string: the newline separator.
+        Defaults to ``os.linesep``.
     :type filename: str
     :type ordered: bool
     :type no_links: bool
@@ -162,7 +170,7 @@ def build_toc(filename: str,
     if not skip_lines >= 0:
         raise ValueError
 
-    toc = str()
+    toc = list()
     header_type_counter = dict()
     header_type_curr = 0
     header_type_prev = 0
@@ -285,8 +293,8 @@ def build_toc(filename: str,
                     header, ordered, no_links, index, parser, list_marker)
 
                 # Save the TOC line with the indentation.
-                toc += build_toc_line(toc_line_no_indent,
-                                      no_of_indentation_spaces_curr) + newline_string
+                toc.append(build_toc_line(toc_line_no_indent,
+                                          no_of_indentation_spaces_curr) + newline_string)
 
                 header_type_prev = header_type_curr
 
@@ -300,6 +308,8 @@ def build_toc(filename: str,
 
     f.close()
 
+    toc = ''.join(toc)
+
     return toc
 
 
@@ -336,6 +346,8 @@ def build_multiple_tocs(filenames: list,
          Defaults to ``-``.
     :parameter constant_ordered_list: use a single integer
         as list marker. This sets ordered to ``True``.
+    :parameter newline_string: the newline separator.
+        Defaults to ``os.linesep``.
     :type filenames: list
     :type ordered: bool
     :type no_links: bool
@@ -345,6 +357,7 @@ def build_multiple_tocs(filenames: list,
     :type list_marker: str
     :type skip_lines: int
     :type constant_ordered_list: bool
+    :type newline_string: str
     :returns: toc_struct, the corresponding table of contents for each input
          file.
     :rtype: list
@@ -683,7 +696,7 @@ def remove_html_tags(line: str, parser: str = 'github') -> str:
 
 
 def remove_emphasis(line: str, parser: str = 'github') -> list:
-    r"""Remove emphasis.
+    r"""Remove markdown emphasis.
 
     :parameter line: a string.
     :parameter parser: decides rules on how to find delimiters.
@@ -693,22 +706,21 @@ def remove_emphasis(line: str, parser: str = 'github') -> list:
     :returns: the input line without emphasis.
     :rtype: str
     :raises: a built-in exception.
+
+    .. note:: Backslashes are preserved.
     """
     if parser in ['github', 'cmark', 'gitlab', 'commonmarker', 'goldmark', 'redcarpet']:
-
         mem = None
-        refmap = None
+        refmap = references_h._cmarkCmarkReferenceMap()
 
-        content = cmark._cmarkCmarkChunk(line, len(line), 0)
-        subj = cmark._cmark_Subject(input=line)
-        cmark._cmark_subject_from_buf(mem, 0, 0, subj, content, refmap)
+        parent = node_h._cmarkCmarkNode()
+        parent.data = line
+        parent.length = len(line)
+        parent.start_line = 0
+        parent.start_column = 0
+        parent.internal_offset = 1
 
-        while not cmark._cmark_is_eof(subj):
-            cmark._cmark_parse_inline(subj)
-
-        ignore = list()
-        # When we hit the end of the input, we call the process emphasis procedure (see below), with stack_bottom = NULL.
-        cmark._cmark_process_emphasis(subj, None, ignore)
+        ignore = inlines_c._cmark_cmark_parse_inlines(mem, parent, refmap, 0)
         line = filter_indices_from_line(line, ignore)
 
     elif parser in ['redcarpet']:
@@ -738,18 +750,20 @@ def filter_indices_from_line(line: str, ranges: list) -> str:
     ll = [item for e in rng for item in e]
 
     s = sorted(ll)
-    final = str()
+    final = list()
     i = 0
     while i < len(line):
         if i not in s:
-            final += line[i]
+            final.append(line[i])
         i += 1
 
+    final = ''.join(final)
+
     return final
 
 
 def build_anchor_link(header_text_trimmed: str,
-                      header_duplicate_counter: str,
+                      header_duplicate_counter: dict,
                       parser: str = 'github') -> str:
     r"""Apply the specified slug rule to build the anchor link.
 
@@ -768,6 +782,9 @@ def build_anchor_link(header_text_trimmed: str,
     :rtype: str
     :raises: a built-in exception.
 
+    .. note: license B applies for the github part.
+        See docs/copyright_license.rst
+
     .. note: license A applies for the redcarpet part.
         See docs/copyright_license.rst
     """
@@ -814,7 +831,7 @@ def build_anchor_link(header_text_trimmed: str,
         header_text_trimmed_len = len(header_text_trimmed)
         inserted = 0
         stripped = 0
-        header_text_trimmed_middle_stage = ''
+        header_text_trimmed_middle_stage = list()
         for i in range(0, header_text_trimmed_len):
             if header_text_trimmed[i] == '<':
                 while i < header_text_trimmed_len and header_text_trimmed[
@@ -829,14 +846,15 @@ def build_anchor_link(header_text_trimmed: str,
             elif not generic._isascii(header_text_trimmed[i]) or STRIPPED.find(
                     header_text_trimmed[i]) != -1:
                 if inserted and not stripped:
-                    header_text_trimmed_middle_stage += '-'
+                    header_text_trimmed_middle_stage.append('-')
                 stripped = 1
             else:
-                header_text_trimmed_middle_stage += header_text_trimmed[
-                    i].lower()
+                header_text_trimmed_middle_stage.append(header_text_trimmed[
+                    i].lower())
                 stripped = 0
                 inserted += 1
 
+        header_text_trimmed_middle_stage = ''.join(header_text_trimmed_middle_stage)
         if stripped > 0 and inserted > 0:
             header_text_trimmed_middle_stage = header_text_trimmed_middle_stage[
                 0:-1]
@@ -906,8 +924,11 @@ def get_atx_heading(line: str,
     :raises: GithubEmptyLinkLabel or GithubOverflowCharsLinkLabel or a
          built-in exception.
 
-    .. note:: license B applies for the github part and license A for redcarpet part.
-         See docs/markdown_specification.rst
+    .. note: license A applies for the redcarpet part.
+         See docs/copyright_license.rst
+
+    .. note:: license B applies for the github part.
+         See docs/copyright_license.rst
     """
     if not keep_header_levels >= 1:
         raise ValueError
@@ -929,6 +950,7 @@ def get_atx_heading(line: str,
             if len(subl) == 0 or subl[0] == '\u005c':
                 continue
 
+            # Preceding.
             i = 0
             while i < len(subl) and subl[i] == ' ' and i <= md_parser['github'][
                     'header']['max space indentation']:
@@ -936,6 +958,7 @@ def get_atx_heading(line: str,
             if i > md_parser['github']['header']['max space indentation']:
                 continue
 
+            # ATX characters.
             offset = i
             while i < len(subl) and subl[i] == '#' and i <= md_parser['github'][
                     'header']['max levels'] + offset:
@@ -946,21 +969,37 @@ def get_atx_heading(line: str,
 
             current_headers = i - offset
 
-            # Include special cases for l endings which should not be
+            # At this moment GFM is still at version 0.29
+            # while cmark is at 0.30. There are subtle differences
+            # such as this one. Assume gitlab is on par with 0.30.
+            if parser in ['github', 'commonmarker']:
+                # GFM 0.29 and cmark 0.29.
+                spaces = [' ']
+            else:
+                # cmark 0.30.
+                spaces = [' ', '\u0009']
+
+            # Include special cases for line endings which should not be
             # discarded as non-ATX headers.
-            if i < len(subl) and (subl[i] != ' ' and subl[i] != '\u000a'
-                                  and subl[i] != '\u000d'):
+            if i < len(subl) and subl[i] not in spaces + ['\u000a', '\u000d']:
                 continue
 
             i += 1
+
             # Exclude leading whitespaces after the ATX header identifier.
-            while i < len(subl) and subl[i] == ' ':
+            # [0.29]:
+            #   The opening sequence of `#` characters must be followed
+            #   by a space or by the end of line.
+            # [0.30]:
+            #   The opening sequence of `#` characters must be followed
+            #   by spaces or tabs, or by the end of line.
+            while i < len(subl) and subl[i] in spaces:
                 i += 1
 
             # An algorithm to find the start and the end of the closing sequence (cs).
             # The closing sequence includes all the significant part of the
             # string. This algorithm has a complexity of O(n) with n being the
-            # length of the l.
+            # length of the line.
             cs_start = i
             cs_end = cs_start
             # subl_prime =~ subl'.
@@ -972,7 +1011,7 @@ def get_atx_heading(line: str,
             hash_round_start = i
 
             # Ignore all characters after newlines and carrage returns which
-            # are not at the end of the l.
+            # are not at the end of the line.
             # See the two CRLF marker tests.
             crlf_marker = 0
             stripped_crlf = False
@@ -994,7 +1033,7 @@ def get_atx_heading(line: str,
 
             # Cut spaces and hashes.
             while go and i < len_subl - cs_start:
-                if (subl_prime[i] not in [' ', '#']
+                if (subl_prime[i] not in spaces + ['#']
                         or hash_char_rounds > 1):
                     if i > hash_round_start and hash_char_rounds > 0:
                         cs_end = len_subl - hash_round_start
@@ -1002,7 +1041,7 @@ def get_atx_heading(line: str,
                         cs_end = len_subl - i
                     go = False
                 if go:
-                    while subl_prime[i] == ' ':
+                    while subl_prime[i] in spaces:
                         i += 1
                     hash_round_start = i
                     while subl_prime[i] == '#':
@@ -1215,8 +1254,22 @@ def is_opening_code_fence(line: str, parser: str = 'github'):
                 i += 1
 
             info_string = line[info_string_start:len(line)]
+            # [0.29]
+            #   The line with the opening code fence may optionally contain
+            #   some text following the code fence; this is trimmed of leading
+            #   and trailing whitespace and called the info string.
+            # [0.30]
+            #   The line with the opening code fence may optionally contain
+            #   some text following the code fence; this is trimmed of leading
+            #   and trailing spaces or tabs and called the info string.
+            # This does not change the end result of this function.
+            if parser in ['github', 'commonmarker']:
+                spaces = ' '
+            else:
+                spaces = ' \u0009'
+            info_string = info_string.strip(spaces)
 
-        # Backticks or tildes in info string are explicitly forbidden
+        # Backticks or tildes in info strings are explicitly forbidden
         # in a backtick-opened code fence, for Commonmark 0.29.
         # This also solves example 107 [Commonmark 0.28]. See:
         # https://github.github.com/gfm/#example-107
@@ -1270,7 +1323,7 @@ def is_closing_code_fence(line: str,
             return False
 
         # Additional security.
-        fence = fence.rstrip('\n').rstrip(' ')
+        fence = fence.rstrip(' \n')
         # Check that all fence characters are equal.
         if fence != len(fence) * fence[0]:
             return False
@@ -1292,7 +1345,19 @@ def is_closing_code_fence(line: str,
         if not line.startswith(fence):
             return False
 
-        line = line.rstrip('\n').rstrip(' ')
+        # [0.29]
+        #   The closing code fence may be indented up to three spaces, and may be
+        #   followed only by spaces, which are ignored.
+        # [0.30]
+        #   The closing code fence may be preceded by up to three spaces of
+        #   indentation, and may be followed only by spaces or tabs,
+        #   which are ignored.
+        if parser in ['github', 'commonmarker']:
+            spaces = ' '
+        else:
+            spaces = ' \u0009'
+        line = line.rstrip(spaces + '\n')
+
         # Solves:
         # example 93 -> 94 [Commonmark 0.28]
         # example 94 -> 95 [Commonmark 0.29]
